--BEHAVIOUR TREE LIB VERSION 0.1--
      --AUTHOR: pedrosgali--

--Basic library for building behaviour trees consisting of decision nodes.
--Each node can add multiple branches leading to other nodes.
--Branches can have three states, PASS, FAIL or RUNNING.
--PASS or FAIL get passed up the tree to the parent node as boolean true or false.
--RUNNING will call the current branch every tick until it returns PASS or FAIL.

local node = {}

node.base = {}

function node.base:extend()
  return setmetatable({}, {__index = self})
end

function node.base:new(...)
  local inst = setmetatable({branches = {},}, {__index = self})
  return inst, inst:init(...), inst:reset()
end

function node.base:init(...) end

function node.base.update(...) end

function node.base:reset()
  self.currentBranch = 1
  if self.branches ~= nil then
    for i = 1, #self.branches do
      self.branches[i]:reset()
    end
  end
end

function node.base:add(n)
  table.insert(self.branches, n)
  self.currentBranch = 1
end

--Loop node, a standard loop that will run the AI.
--It will go through each added branch until it reaches the end then start again.
--All AI's will need a loop node as the starting node otherwise the tree will only run once.

node.loop = node.base:extend()

function node.loop:update(...)
  local state = self.branches[self.currentBranch]:update(...)
  if state == true or state == false then
    self.currentBranch = self.currentBranch + 1
  end
  if self.currentBranch > #self.branches then
    self:reset()
    return true
  end
end

--Invert will return fail for a pass and pass for a fail,

node.invert = node.base:extend()

function node.invert:update(...)
  local state = self.branches[self.currentBranch]:update(...)
  if state == true or state == false then
    return not state
  end
  return "RUNNING"
end

--The sequence node is an AND operator requiring all branches to pass
--before it will return a pass.  Any fail result will return fail instantly.

node.sequence = node.base:extend()

function node.sequence:update(...)
  local state = self.branches[self.currentBranch]:update(...)
  if state == true then
    self.currentBranch = self.currentBranch + 1
  elseif state ==false then
    return false
  end
  if self.currentBranch > #self.branches then
    return true
  end
  return "RUNNING"
end

--The selector node is the OR operator, any pass will return pass instantly

node.selector = node.base:extend()

function node.selector:update(...)
  local state = self.branches[self.currentBranch]:update(...)
  if state == false then
    self.currentBranch = self.currentBranch + 1
  elseif state == true then
    return true
  end
  if self.currentBranch > #self.branches then
    return false
  end
  return "RUNNING"
end

return node