local cam = engine.class.base:extend()

function cam:init(x, y, w, h, scale)
  self:setPosition(x, y)
  self:setSize(w, h)
  self:setScale(scale)
  self:setTarget(0)
end

function cam:setPosition(x, y)
  self.x, self.y = x, y
  self.target = 0
end

function cam:getPosition()
  return self.x, self.y
end

function cam:setSize(w, h)
  self.w, self.h = w, h
  self.canvas = love.graphics.newCanvas(w, h)
  self.starfield = {
    love.graphics.newCanvas(w, h),
    love.graphics.newCanvas(w, h),
  }
end

function cam:setTarget(target)
  self.target = target
end

function cam:getInWorldSize()
  return self.w * (1 / self.scale), self.h * (1 / self.scale)
end

function cam:setScale(scale)
  self.scale = scale
end

function cam:getDimensions()
  return self.x, self.y, self.w, self.h, self.scale
end

function cam:getBounds()
  local fl = math.floor
  local x, y, w, h, s = self:getDimensions()
  w = w * (1 / s)
  h = h * (1 / s)
  local xmin = fl(x - (w / 2))
  local ymin = fl(y - (h / 2))
  return xmin, ymin, w, h, s
end

function cam:getScreenPos(x, y)
  local minx, miny, w, h, s = self:getBounds()
  local dx = (x - self.x)
  local dy = (y - self.y)
  local rx = ((self.w / 2) + (dx * s))
  local ry = ((self.h / 2) + (dy * s))
  return rx, ry
end

function cam:getMapPos(x, y)
  local xMin, yMin, w, h, s = self:getBounds()
  s = 1 / s
	local retx = math.floor(xMin + (x * s))
	local rety = math.floor(yMin + (y * s))
	return retx, rety
end

function cam:checkControls()
  local isDown = love.keyboard.isDown
  local mult = (5 * (1 / game.world.cameras.main.scale))
  if isDown("w") then
    local x, y = game.world.cameras.main:getPosition()
    game.world.cameras.main:setPosition(x, y - mult)
  end
  if isDown("s") then
    local x, y = game.world.cameras.main:getPosition()
    game.world.cameras.main:setPosition(x, y + mult)
  end
  if isDown("a") then
    local x, y = game.world.cameras.main:getPosition()
    game.world.cameras.main:setPosition(x - mult, y)
  end
  if isDown("d") then
    local x, y = game.world.cameras.main:getPosition()
    game.world.cameras.main:setPosition(x + mult, y)
  end
end

function cam:update(dt, world)
  if self.target ~= 0 then
    local tr = world:getObjectData(self.target, "transform")
    self.x, self.y = tr.x, tr.y
  end
  local mult = 10 * (1 / self.scale)
  engine.shaders.starfield:send("iMapPosX", self.x / mult)
  engine.shaders.starfield:send("iMapPosY", self.y / mult)
  engine.shaders.starfield:send("iResolutionX", engine.screenWidth)
  engine.shaders.starfield:send("iResolutionY", engine.screenHeight)
  engine.shaders.starfield:send("iScale", math.max(self.scale / 500, 1))
  self:drawStars(1)
  self:checkControls()
end

function cam:drawDebug()
  local h = font:getHeight()
  local mapx, mapy = self:getMapPos(engine.mousex, engine.mousey)
  love.graphics.setColor(0,255,0,255)
  love.graphics.print("FPS: "..love.timer.getFPS(), 1, 1)
  love.graphics.print("Entity Count: "..#self.renderList, 1, h)
  love.graphics.print("Zoom Level: "..self.scale, 1, h * 2)
  love.graphics.print("Game speed: "..game.world.speed.."x", 1, h * 3)
  love.graphics.print("Map X: "..mapx, 1, h * 4)
  love.graphics.print("Map Y: "..mapy, 1, h * 5)
end

function cam:drawStars(i)
  local lastCanvas = love.graphics.getCanvas()
  love.graphics.setCanvas(self.starfield[i])
  love.graphics.clear(0,0,0,255)
  love.graphics.setColor(255,255,255,255)
  love.graphics.setShader(engine.shaders.starfield)
  love.graphics.draw(self.canvas, 0, 0)
  love.graphics.setShader()
  love.graphics.setCanvas(lastCanvas)
end

function cam:draw(world)
  local lastCanvas = love.graphics.getCanvas()
  love.graphics.setCanvas(self.canvas)
  love.graphics.setColor(255,255,255,255)
  love.graphics.draw(self.starfield[1], 0, 0)
  if self.renderList == nil then return end
  if self.renderList[1] == nil then return end
  for i = 1, #self.renderList do
    local id = self.renderList[i]
    local data = world.objects[id]
    data.renderer.x, data.renderer.y = self:getScreenPos(data.transform.x, data.transform.y)
    data.renderer.screenScale = self.scale
    if data.type == "planet" or data.type == "moon" then
      engine.behaviour["planet_renderer"]:draw(data)
    else
      engine.behaviour["renderer"]:draw(data)
    end
  end
  self:drawDebug()
  love.graphics.setCanvas()
end

return cam