local world = engine.class.base:extend()

function world:init()
  self.speed = 1
  self.isPaused = false
  self.objects = {}
  self.players = {}
  self.cameras = {}
  self.typeLists = {}
  self.cullCount = 1
  self.cullTimer = 4
  self:addCamera("main", engine.system.camera:new(0, 0, engine.screenWidth, engine.screenHeight, .05))
  dbg.add("Loading Start...", "WORLD LOAD")
  self:loadSystems()
end

function world:addCamera(key, cam)
  self.cameras[key] = cam
end

function world:removeCamera(key)
  self.cameras[key] = nil
end

function world:nextID(t)
  for i = 1, #t do
    if t[i] == nil then
      return i
    end
  end
  return #t + 1
end

function world:loadSystems()
  self.systemLists = {}
  self.renderLists = {}
  for k, v in pairs(engine.behaviour) do
    dbg.add("Loading behaviour: "..k, "WORLD LOAD")
    self.systemLists[k] = {}
  end
end

function world:checkSysRequirements(sysTab, objTab)
  for i = 1, #sysTab do
    local reqFound = false
    for k, v in pairs(objTab) do
      if k == sysTab[i] then
        reqFound = true
        break
      end
    end
    if not reqFound then return false end
  end
  return true
end

function world:registerSystems(id, data)
  for k, v in pairs(engine.behaviour) do
    if type(v) == "table" then
      if self:checkSysRequirements(v.requirements, data) then
        self.objects[id] = engine.behaviour[k]:startup(self.objects[id])
        self.systemLists[k][#self.systemLists[k] + 1 or 1] = id
      end
    end
  end
end

function world:unRegisterSystems(id)
  for k, v in pairs(self.systemLists) do
    for i = 1, #self.systemLists[k] do
      if self.systemLists[k][i] == id then
        table.remove(self.systemLists[k], i)
        break
      end
    end
  end
end

function world:addComponent(id, comp, ...)
  self:unRegisterSystems(id)
  self.objects[id][comp] = engine.component[comp].new(...)
  if engine.component[comp].init ~= nil then
    engine.component[comp].init(id, self.objects[id][comp])
  end
  self:registerSystems(id, self.objects[id])
end

function world:addBuilding(id, b, ...)
  table.insert(self.objects[id].colony.building, 1, engine.buildings[b].new(...))
  engine.buildings[b].init(self.objects[id], self.objects[id].colony.building[1])
end

function world:addColonyGoods(id, item, amt)
  if self.objects[id].colony ~= nil then
    self.objects[id].colony.inventory[item].amt = self.objects[id].colony.inventory[item].amt + amt
    dbg.add(amt.." "..item.." added to colony at "..id, "ADD GOODS")
  end
end

function world:addPlayer(name, cash, home, ai)
  table.insert(self.players, engine.system.player:new(name, cash, home, ai))
  local id = #self.players
  self.players[id].id = id
  if ai ~= "none" then
    self.players[id].aiTree = engine.AI[ai]:new()
  end
  dbg.add("Player added: "..id.." "..name.." at "..home, "WORLD SPAWN")
  return id
end

function world:add(data)
  local id = self:nextID(self.objects)
  self.objects[id] = data
  self.objects[id].id = id
  --dbg.add("Entity added: "..id, "WORLD SPAWN")
  self:registerSystems(id, data)
  if self.typeLists[data.type] == nil then
    self.typeLists[data.type] = {}
  end
  table.insert(self.typeLists[data.type], id)
  return id
end

function world:remove(id)
  self:unRegisterSystems(id)
  self.objects[id] = "none"
end

function world:getObjectData(id, key)
  if key == nil then
    return self.objects[id]
  elseif self.objects[id][key] ~= nil then
    return self.objects[id][key]
  end
end

function world:getObjectDistance(id, tId)
  local dx, dy = self:getObjectVector(id, tId)
  local r = self:getObjectData(id, "renderer")
  local t = self:getObjectData(tId, "transform")
  return math.sqrt((dx * dx) + (dy * dy)) - ((r.w * t.scale) / 3)
end

function world:getObjectVector(id, tId)
  local p = self:getObjectData(id, "transform")
  local t = self:getObjectData(tId, "transform")
  return t.x - p.x, t.y - p.y
end

function world:deselectEntities(playerId)
  for i = 1, #self.objects do
    self.objects[i].renderer.isSelected = false
  end
  self.players[playerId].selected = {}
end

function world:selectEntity(id)
  self.objects[id].renderer.isSelected = true
end

function world:selectShips(playerId, minx, miny, maxx, maxy)
  self:deselectEntities(playerId)
  local ships = self.players[playerId].ships
  local count = 1
  if ships[1] ~= nil then
    for i = 1, #ships do
      local tr = self:getObjectData(ships[i], "transform")
      if tr.x >= minx and tr.x <= maxx then
        if tr.y >= miny and tr.y <= maxy then
          self.players[playerId]:selectShip(ships[i])
          self:selectEntity(ships[i])
          count = count + 1
        end
      end
    end
  end
  for i = 1, #self.players[playerId].selected do
    dbg.add("Ship "..self.players[playerId].selected[i].." selected.", "SELECT")
  end
end

function world:checkClick(x, y, b)
  local rList = self.cameras.main.renderList
  local mx, my = self.cameras.main:getMapPos(x, y)
  for i = 1, #rList do
    local t = self:getObjectData(rList[i], "transform")
    local r = self:getObjectData(rList[i], "renderer")
    local rad = (r.w * t.scale) / 2
    local dx, dy = t.x - mx, t.y - my
    if math.abs(dx) < rad and math.abs(dy) < rad then
      return rList[i]
    end
  end
end

function world:cullLists()
  local fl = math.floor
  local ins = table.insert
  for c, _ in pairs(self.cameras) do
    self.cameras[c].renderList = {}
    self.cameras[c].xmin, self.cameras[c].ymin, self.cameras[c].scaleWidth, self.cameras[c].scaleHeight = self.cameras[c]:getBounds()
  end
  for i = 1, #self.systemLists["renderer"] do
    local id = self.systemLists["renderer"][i]
    local data = self:getObjectData(id)
    for c, _ in pairs(self.cameras) do
      local x1, y1 = self.cameras[c].xmin, self.cameras[c].ymin
      local w1, h1 = self.cameras[c].scaleWidth, self.cameras[c].scaleHeight
      local t = data.transform
      local r = data.renderer
      local w2 = r.w * t.scale
      local h2 = r.h * t.scale
      local x2 = t.x - fl(w2 / 2)
      local y2 = t.y - fl(h2 / 2)
      if x1 < x2 + w2 and x2 < x1 + w1 and y1 < y2 + h2 and y2 < y1 + h1 then
        ins(self.cameras[c].renderList, id)
      end
    end
  end
end

function world:wheelMoved(x, y, s)
  local sc = self.cameras.main.scale
  if s < 0 then
    s = -1 * sc / 2
    self.cameras.main:setScale(math.max(sc + s, .00001))
  else
    s = 1 * sc / 2
    self.cameras.main:setScale(math.min(sc + s, 50))
  end
end

function world:update(dt)
  if not self.isPaused then
    for k, v in pairs(self.systemLists) do
      for i = 1, #self.systemLists[k] do
        for s = 1, self.speed do
          engine.behaviour[k]:update(dt, self.objects[self.systemLists[k][i]])
        end
      end
    end
  end
  for k, v in pairs(self.cameras) do
    self.currentCamera = k
    self.cameras[k]:update(dt, self)
  end
  for i = 1, #self.players do
    if self.players[i].aiTree ~= nil then
      self.players[i].aiTree:update(self.players[i])
    end
  end
  if self.dragSelect then
    self.mapx2, self.mapy2 = game.world.cameras.main:getMapPos(engine.mousex, engine.mousey)
  end
  self.cullCount = self.cullCount + 1
  if self.cullCount >= self.cullTimer then
    self:cullLists()
    self.cullCount = 0
  end
end

function world:draw()
  for k, c in pairs(self.cameras) do
    self.currentCamera = k
    c:draw(self)
  end
  love.graphics.setCanvas()
  engine.ui.colour.set("white")
  love.graphics.draw(self.cameras.main.canvas, 0, 0)
  if self.dragSelect then
    engine.ui.colour.set("green")
    local x, y = self.cameras.main:getScreenPos(math.min(self.mapx, self.mapx2), math.min(self.mapy, self.mapy2))
    local x2, y2 = self.cameras.main:getScreenPos(math.max(self.mapx, self.mapx2), math.max(self.mapy, self.mapy2))
    local w, h = x2 - x, y2 - y
    love.graphics.rectangle("line", x, y, w, h)
  end
end

return world