local win = engine.ui.container:extend()

function win:init()
  self.isVisible = true
  self.isHovered = false
  self.isMoveable = false
  self.isGrabbed = false
  self.isSelected = true
  self.isDirty = true
  self.isTiled = true
  self.canvas = love.graphics.newCanvas(8, 8)
  self:internalSetup()
end

function win:internalSetup()
  self.type = "window"
  self.headerHeight = 32
  self.xOffset = 0
  self.yOffset = self.headerHeight
  self.scrollY = 0
  self:setMoveable(true)
end

function win:setHeaderHeight(h)
  self.headerHeight = h
end

function win:add(key, object, ...)
  if self.objects == nil then self.objects = {} end
  if self.drawOrder == nil then self.drawOrder = {} end
  self.objects[key] = object
  self.objects[key].key = key
  if self.win == nil then
    self.objects[key].win = self
  else
    self.objects[key].win = self.win
  end
  self.objects[key].box = self
  self.objects[key].boxId = self.key
  self.objects[key]:setup(...)
  if object.y + object.h + 5 > self.h then
    self:setSize(self.w, object.y + object.h + 5)
  end
  table.insert(self.drawOrder, 1, key)
end

function win:onClose() end

function win:mousePressed(x, y, b)
  if self:checkHovered(x, y) then
    self:deselectAll()
    if self.isVisible then
      if self.drawOrder ~= nil then
        if self.type == "window" then
          x = x - self.x
          y = y - self.y
        else
          local xPos, yPos = self:getScaledPosition()
          x, y = x - xPos, y - yPos
        end
        for i = 1, #self.drawOrder do
          local scrollOffset = 0
          if self.drawOrder[i] ~= "scroll_bar" then
            scrollOffset = self.scrollY
          end
          if self.objects[self.drawOrder[i]]:mousePressed(x, y + scrollOffset, b) then
            self.stretch = false
            self.isGrabbed = false
            return true
          end
        end
      end
      local w, h = self:getSize()
      if (x < 5 or x > w - 5) or y < 5 or y > h - 5 then
        self.stretch = true
      else
        self.isGrabbed = true
        self.isDirty = true
      end
      self.grabx = x
      self.graby = y
      return true
    end
  end
  return false
end

function win:updateMaster(dt, mx, my)
  if mx == nil then
    mx, my = engine.mousex, engine.mousey
  end
  self:checkHovered(mx, my)
  if self.drawOrder ~= nil then
    local scrollY = self.scrollY or 0
    local x, y = mx, my
    x = x - self.x
    y = y - self.y
    for i = 1, #self.drawOrder do
      if self.drawOrder[i] ~= "scroll_bar" then
        y = y + (scrollY * self.yScale)
      end
      self.objects[self.drawOrder[i]]:updateMaster(dt, x, y)
      if self.objects[self.drawOrder[i]].y < scrollY then
        self.objects[self.drawOrder[i]].isDirty = false
        self.objects[self.drawOrder[i]].isHovered = false
      end
    end
  end
  if self.isMoveable and self.isGrabbed then
    local newx = engine.mousex - self.grabx
    local newy = engine.mousey - self.graby
    self:setPosition(newx, newy)
  end
  self:update(dt, mx, my)
  if self.isDirty then
    self:drawBackground()
    self:redrawObjects()
  elseif self.drawOrder ~= nil then
    for i = #self.drawOrder, 1, -1 do
      self.objects[self.drawOrder[i]]:draw()
    end
  end
end

function win:drawLabel()
  engine.ui.colour.set(self.tCol)
  local tw = font:getWidth(self.label)
  local th = font:getHeight()
  local tx = math.floor((self.w - tw) / 2)
  local ty = math.floor((self.headerHeight - th) / 2 - 1)
  love.graphics.print(self.label, tx, ty)
end

function win:drawShade()
  local topLeft = {x = 0, y = 0}
  local topRight = {x = self.w, y = 0}
  local bottomLeft = {x = 0, y = self.h}
  local bottomRight = {x = self.w, y = self.h}
  love.graphics.setColor(engine.system.vars.user.uiColour)
  love.graphics.line(topLeft.x, topLeft.y, topRight.x, topRight.y)
  love.graphics.line(topLeft.x + 1, topLeft.y + 1, topRight.x - 1, topRight.y + 1)
  love.graphics.line(topLeft.x, topLeft.y, bottomLeft.x, bottomLeft.y)
  love.graphics.line(topLeft.x + 1, topLeft.y + 1, bottomLeft.x - 1, bottomLeft.y - 1)
  engine.ui.colour.set(engine.ui.colour.shade("black"))
  love.graphics.line(bottomRight.x, bottomRight.y, bottomLeft.x, bottomLeft.y)
  love.graphics.line(bottomRight.x - 1, bottomRight.y - 1, bottomLeft.x + 1, bottomLeft.y - 1)
  love.graphics.line(bottomRight.x, bottomRight.y, topRight.x, topRight.y)
  love.graphics.line(bottomRight.x - 1, bottomRight.y - 1, topRight.x - 1, topRight.y - 1)
end

function win:drawBackground()
  local lastCanvas = self:getCanvas()
  if self.headerHeight == nil then self.headerHeight = 20 end
  if engine.assets[self.image] ~= nil then
    self:tileBackground()
    self:drawLabel()
  else
    engine.ui.colour.set(self.bCol)
    love.graphics.rectangle("fill", 0, 0, self.w, self.h)
    engine.ui.colour.set(self.hCol)
    love.graphics.rectangle("fill", 0, 0, self.w, self.headerHeight)
    self:drawLabel()
    self:drawShade()
  end
  love.graphics.setCanvas(lastCanvas)
  self.isDirty = false
end

function win:draw()
  if self.isVisible then
    if self.isDirty then
      self:drawBackground()
      self:redrawObjects()
    end
    if self.drawOrder ~= nil then
      for i = #self.drawOrder, 1, -1 do
        self.objects[self.drawOrder[i]]:draw()
      end
    end
    engine.ui.colour.set("white")
    love.graphics.draw(self.canvas, self.x, self.y)
  end
end

return win